/********************************************************************************/
/*                                                                              */
/*  File:   WinMain.cpp                                                         */
/*  Auth:   Ben Kenwright                                                       */
/*  Email:  bkenwright@xbdev.net                                                */
/*  Url:    www.xbdev.net                                                       */
/*  Date:   25/12/06 (xmas)                                                     */
/*                                                                              */
/********************************************************************************/

#include <windows.h>
#include <stdio.h>		// strcat(..)

#include "resource.h"
#include "XFile.h"


HWND g_hwdTextBox = NULL;

// Forward Function Declarations
int __stdcall DialogProc(HWND hWnd, UINT uMsg, WPARAM wParam, LPARAM lParam);


/********************************************************************************/
/*                                                                              */
/*  WinMain(..)                                                                 */
/*  Program Entry Point                                                         */
/*                                                                              */
/********************************************************************************/
int __stdcall WinMain(HINSTANCE hInst, HINSTANCE hPrev, LPSTR szCmdLine, int nCmdShow)
{
	HWND hDialog = 0;

    hDialog = CreateDialog (hInst, 
                            MAKEINTRESOURCE (IDD_DIALOG1), 
                            0, 
                            DialogProc);

    if (!hDialog)
    {
		char buf [100];
		wsprintf (buf, "Error x%x", GetLastError ());
		MessageBox (0, buf, "CreateDialog", MB_ICONEXCLAMATION | MB_OK);
		return 1;
    }

	UpdateWindow(hDialog);
	ShowWindow(hDialog, nCmdShow);

	g_hwdTextBox = GetDlgItem(hDialog, IDC_TEXTBOX);

    MSG  msg;
    int status;
    while ((status = GetMessage (& msg, 0, 0, 0)) != 0)
    {
        if (status == -1)
            return -1;
        if (!IsDialogMessage (hDialog, & msg))
        {
            TranslateMessage ( & msg );
            DispatchMessage ( & msg );
        }
    }

    return (int)msg.wParam;

}// End WinMain(..)


/********************************************************************************/
/*                                                                              */
/*  DialogProc(..)                                                              */
/*  Call back function - all messages get send to here from our dialog          */
/*                                                                              */
/********************************************************************************/


int __stdcall DialogProc (HWND hWnd, UINT uMsg, WPARAM wParam, LPARAM lParam)
{
	OPENFILENAME ofn;
	char Filename[MAX_PATH];


    switch (uMsg)
	{
		case WM_COMMAND:
		{
			switch(LOWORD(wParam)) 
			{
				case IDC_OPENFILE:
				{
					// Clear list
					SendMessage(GetDlgItem(hWnd, IDC_TEXTBOX), LB_RESETCONTENT, 0, 0);

					// Get filename to open
					Filename[0] = 0;
					ZeroMemory(&ofn, sizeof(OPENFILENAME));
					ofn.lStructSize = sizeof(OPENFILENAME);
					ofn.nMaxFile = MAX_PATH;
					ofn.nMaxFileTitle = MAX_PATH;
					ofn.Flags = OFN_HIDEREADONLY | OFN_CREATEPROMPT | OFN_OVERWRITEPROMPT;
					ofn.hwndOwner   = hWnd;
					ofn.lpstrFile   = Filename;
					ofn.lpstrTitle  = "Load and Parse .X File";
					ofn.lpstrFilter = ".X Files (*.x)\0*.x\0All Files (*.*)\0*.*\0\0";
					ofn.lpstrDefExt = "x";
					if(!GetOpenFileName(&ofn))
					{
						return 0;
					}

					// Parse the .X file and dump data to the text box
					ParseXFile(Filename);
				}
				break;
			}
			break;
		}
		break;

		case WM_SIZE:
		{
			HWND hEdit;
			RECT rcClient;

			GetClientRect(hWnd, &rcClient);

			hEdit = GetDlgItem(hWnd, IDC_TEXTBOX);
			const int topBorder = 50;
			const int bottomBorder = 10;
			const int leftBorder = 10;
			const int width  = rcClient.right - 2*leftBorder;
			const int height = rcClient.bottom - topBorder - bottomBorder;
			SetWindowPos(	hEdit, 
							NULL, 
							leftBorder,
							topBorder,
							width, 
							height, 
							SWP_NOZORDER);
		}
		break;

		case WM_DESTROY:
		{
			PostQuitMessage(0);
			return TRUE;
		}
		break;

		case WM_CLOSE:
		{
			DestroyWindow (hWnd);
			return TRUE;
		}
		break;
    }
    return FALSE;
}



/********************************************************************************/
/*                                                                              */
/*  AddToTextBox(..)                                                            */
/*  Any text we want to add to the text box we call this function.              */
/*                                                                              */
/********************************************************************************/
void AddToTextBox(char* str, int indent /*=0*/ )
{
	if (strlen(str) > 1023 )
	{
		// overflow error
		return;
	}

	char buf[1024]={'\0'};

	// Error checking
	if(str==NULL)
	{
		return;
	}

	// Build text to add to list
	memset(buf, ' ', indent);
	buf[indent] = 0;
	strcat(buf, str);

	int strLen = (int)SendMessage(g_hwdTextBox, WM_GETTEXTLENGTH, 0, 0) + 1;
	int strLenNewStr = (int)strLen + (int)strlen(str);
	char *strTemp = new char[strLenNewStr];

	ZeroMemory(strTemp, strLenNewStr);
	SendMessage(g_hwdTextBox, WM_GETTEXT, strLen, (LPARAM)strTemp);
	strcat(strTemp, str);
	SendMessage(g_hwdTextBox, WM_SETTEXT, 0, (LPARAM)strTemp);

	delete[] strTemp;
}// AddToTextBox(..)




